from datetime import datetime
from pandas import read_csv

from marge.config import config
from marge.enumerations import *
from marge.utils import *

# cleans data frame for first pass
# basically, removes any unnecessary columns
def trim(df, model_name):
    keeping_these_columns = config["models"][model_name]["columns"]
    return df[keeping_these_columns]

# gets value for a given key
def get_value(obj, key):
    if isinstance(obj, dict):
        return obj.get(key, None)
    elif hasattr(obj, key):
        return getattr(obj, key)

def simple_has(obj, key):
    return 1 if get_value(obj, key) not in nulls else 0

def has(obj, key):
    return simple_has(obj, "has_" + key) or simple_has(obj, key)

def has_over(obj, key, threshold):
    try:
        value = get_value(obj, key)
        if value in nulls:
            return 0
        elif isinstance(value, int) or isinstance(value, float):
            return 1 if value > threshold else 0
        elif isinstance(value, str) and value.isdigit():
            return 1 if int(value) > threshold else 0
        else:
            return 0
    except Exception as e:
        print("has_over caught error ", e)
        print("with obj", obj)
        print("with key:", key)
        print("with threshold:", threshold)
        raise(e)

def tofloat(i):
    if isinstance(i, float):
        return i
    elif isinstance(i, int):
        return float(int)
    elif isinstance(i, str):
        return float(i.replace(",", ""))

def is_truthy(i):
    return 1 if i in truthies else 0



import json
import unittest

from marge.cleaner import *

class TestCleaner(unittest.TestCase):

    def test_has(self):
        data = { }
        actual = has(data, "enwiki_tile")
        expected = 0
        self.assertEqual(actual, expected)

        data = { "enwiki_tile": "" }
        actual = has(data, "enwiki_tile")
        expected = 0
        self.assertEqual(actual, expected)

        data = { "enwiki_tile": "Spain" }
        actual = has(data, "enwiki_tile")
        expected = 1
        self.assertEqual(actual, expected)

    def test_has_over(self):
        data = { "population": '147629' }
        actual = has_over(data, "population", 1e2)
        expected = 1
        self.assertEqual(actual, expected)



from georich import enrich
from numpy import where
from pandas import DataFrame, read_csv
import pprint
import pickle
from sklearn.linear_model import LinearRegression, LogisticRegression, SGDRegressor
from sklearn.tree import DecisionTreeRegressor

from marge.config import config
from marge.utils import to_dicts

pp = pprint.PrettyPrinter(indent=4)

class Model:

    def __init__(self, name):
        self.config = config["models"][name]

    def convert(self, i):
        return i if isinstance(i, DataFrame) else DataFrame(i)

    # filter out dicts who have None in one of the relevant columns
    def filter_dicts(self, dicts):
        #return [d for d in dicts if all(v not in [None, ""] for v in d.values())]
        results = []
        for d in dicts:
            keep = True
            for key, value in d.items():
                if key in self.config["columns"] and value in [None, ""]:
                    #print("not keeping kv", [key], [value])
                    keep = False
            if keep:
                results.append(d)
        return results

    def load(self):
        with open(self.config["path"], "rb") as f:
            return pickle.load(f)

    def predict(self, inpt):
        dicts = enrich(inpt, new_fields=self.config["columns"], in_memory=True)
        filtered = self.filter_dicts(dicts)
        if filtered:
            df = self.convert(filtered)
            trimmed = self.trim(df)
            df["score"] = self.load().predict(trimmed)
            return df

    def train(self, train_set=None):
        if train_set is None:
            train_set = to_dicts(config["files"]["dirty_training_file"], nrows=2000)
        dicts = enrich(train_set, new_fields=self.config["columns"], in_memory=True)
        filtered = self.filter_dicts(dicts)
        print("filtered")
        #pp.pprint(filtered)
        df = self.convert(filtered)
        model = LinearRegression()
        #model = SGDRegressor()
        #model = LogisticRegression()
        #model = DecisionTreeRegressor()
        X = self.trim(df)
        self.validate_df(X)
        Y = df["correct"]
        print("Y:", Y.dtypes)
        print("Y:", set(Y))
        model.fit(X, Y)
        #pp.pprint(model.coef_)
        pp.pprint(dict(zip(self.config["columns"], [round(n, 2) for n in model.coef_])))
        with open(self.config["path"], "wb") as f:
            pickle.dump(model, f)

    def trim(self, df):
        return df[self.config["columns"]]

    def trim_dicts(self, dicts, columns=None):
        columns = self.config["columns"] + ["correct"]
        return [
            dict([(k, v) for k, v in d.items() if k in columns ])
            for d in dicts
        ]

    def validate_df(self, df):
        print("validating df")
        for name in self.config["columns"]:
            if len(set(df[name])) == 1:
                raise Exception("UH OH: there's only one in the set for " + name)



from csv import DictReader
import json
from itertools import groupby
from numpy import mean
from pandas import DataFrame
import pprint
import unittest

from marge.config import config
from marge.enricher import add_likely_correct, add_median_cooccurrence, combine_fields
from marge.resolver import *
from marge.models import Model
from marge.utils import to_dicts


pp = pprint.PrettyPrinter(indent=4)

class BaseModelTestCase(unittest.TestCase):

    def setUp(self):
        print("setting up TestFirstPass")
        self.m1 = Model("first_pass")
        self.m2 = Model("second_pass")
        self.test_set = to_dicts(config["files"]["dirty_testing_file"], nrows=config["testing"]["nrows"])

    def score(self, test_set, model):
        scores = []
        try:
            grouped = groupby(test_set, key=lambda x: x["order_id"])
        except Exception as e:
            print(e)
            raise(e)
        for order_id, order in grouped:
            num_incorrect = 0
            num_correct = 0
            for feature_id, features in groupby(order, key=lambda x: x["feature_id"]):
                #print("\n\n\nfeature_id:", feature_id)
                #pp.pprint(list(features))
                results = model.predict(list(features))
                if results is not None:
                    rows = [row for row_index, row in results.iterrows()]
                    if rows:
                        max_score = max([row["score"] for row in rows])
                        correct_rows = [row for row in rows if row["correct"]]
                        if correct_rows:
                            correct = correct_rows[0]
                            # do I have to worry about floating point arithmetic??
                            if correct["score"] == max_score:
                                num_correct += 1
                            else:
                                num_incorrect += 1
                        else:
                            num_incorrect += 1

            total = num_correct + num_incorrect
            if total:
                score = float(num_correct) / total
                scores.append(score)
        mean_score = mean(scores)
        print("len(score:", scores)
        self.assertTrue(len(scores) > 10)
        return mean_score


class TestFirstPass(BaseModelTestCase):
    def test_first_pass(self):
        print("\n\n\nstarting test_first_pass")
        test_set = combine_fields(self.test_set, ["order_id", "feature_id"])
        mean_score = self.score(test_set, self.m1)
        print("m1 mean_score:", round(mean_score,2))
        self.assertTrue(mean_score > 0.9)

class TestSecondPass(BaseModelTestCase):
    def test_second_pass(self):
        print("\n\n\nstarting test_second_pass")
        test_set = combine_fields(self.test_set, ["order_id", "feature_id"])
        first_results = to_dicts(self.m1.predict(test_set))
        places = add_likely_correct(first_results)
        places = add_median_cooccurrence(places)
        mean_score = self.score(places, self.m2)
        print("m2 mean_score:", round(mean_score,2))
        self.assertTrue(mean_score > 0.95)



from collections import Counter
from csv import DictReader
from pandas import DataFrame
import json

from marge.config import config
from marge.enumerations import *
from marge.enumerations import incompletes

def clone(obj):
    return json.loads(json.dumps(obj))


def truthify(x):
    return 0 if x in [None, "", False, "false"] else 1

def is_admin(row):

    try:
        if "geonames_feature_class" in row and row["geonames_feature_class"] == "A":
            return 1

        for key in ["admin_level"]:
            if key in row and row[key]:
                return 1
            try:
                if int(row[key]) > 0:
                    return 1
            except:
                return 0

        return 0
    except Exception as e:
        print(row)
        print(e)
        raise(e)

def is_country(row):
    try:
        if int(row['admin_level']) <= 2:
            return 1
        else:
            return 0
    except:
        return 0

def has_pop_over(entity, threshold):
    try:
        if hasattr(entity, "population"):
            if int(entity.population) > threshold:
                return 1
            else:
                return 0
        elif int(entity["population"]) > threshold:
            return 1
        else:
            return 0
    except:
        return 0

def numerify(obj):
    for key, value in obj.items():
        if isinstance(value, str):
            if value == "":
                obj[key] = 0
            else:
                try:
                    obj[key] = float(value)
                except:
                    pass
    return obj


def get_model(model_name):
    with open(MODEL_PATH, "rb") as f:
        return pickle.load(fail)

def to_dicts(inpt, nrows=None):
    dicts = []
    if isinstance(inpt, str):
        print("to_dicts filepath:", inpt)
        with open(inpt) as f:
            reader = DictReader(f, delimiter="\t")
            count = 0
            for row in reader:
                count += 1
                if nrows and count > nrows:
                    break
                dicts.append(row)
    elif isinstance(inpt, DataFrame):
        count = 0
        for index, series in inpt.iterrows():
            count += 1
            if nrows and count > nrows:
                break
            dicts.append(series.to_dict())
    return dicts


def max_by_group(iterator, column_name, group_name):
    try:
        maxes = {}
        for item in iterator:
            gid = item[group_name]
            value = item[column_name]
            if gid not in maxes or value > maxes[gid]:
                maxes[gid] = value
        return maxes
    except Exception as e:
        print("max_by_group failed with", e)

def is_complete(item):
    if isinstance(item, str):
        return item not in ["", "blank", "null", "Null", "NULL", "NONE", "none", "None"]
    elif isinstance(item, float) or isinstance(item, int):
        return True
    elif item != None:
        return True
    else:
        return False



from marge.models import Model
from marge.utils import to_dicts

def resolve(inpt, debug=True):

    print("starting resolve")
    if debug:
        print("\tinpt:", type(inpt))
        print("\tlen(inpt):", len(inpt))
    first_results = Model("first_pass").predict(inpt)
    print("first_results:", first_results)
    second_results = Model("second_pass").predict(to_dicts(first_results))
    return second_results


import yaml
from os.path import join
from os.path import realpath
from os.path import dirname
import os

dir_path = dirname(realpath(__file__))
filepath = join(dir_path,"config.yml")
print("path to config.yml:", filepath)
with open(filepath) as f:
    config = yaml.load(f.read())



"""
marge specific geoenrichment
i.e. stuff that wouldn't make sense in georich
"""
import pickle

from collections import defaultdict, Counter
from config import config
from converter import to_list_of_dicts
from numpy import median
from itertools import groupby
from utils import max_by_group

def combine_fields(places, fields):
    places = to_list_of_dicts(places)
    combined_name = "-".join(fields)
    for place in places:
        place[combined_name] = "-".join([place[field] for field in fields])
    return places

def add_likely_correct(places):

    print("starting add_likely_correct with places:", len(places))
    places = to_list_of_dicts(places)
    print("places converted:", len(places))

    old_fields = places[0].keys()

    if "score" in old_fields and "feature_id" in old_fields:
        maxes = max_by_group(places, "score", "order_id-feature_id")
        for place in places:
            gid = place["order_id-feature_id"]
            prob = place["score"]
            place["likely_correct"] = 1 if prob == maxes[gid] else 0

    print("added likely correct to all places")
    return places


def add_median_cooccurrence(places):

    print("starting add_median_cooccurrence")
    places = to_list_of_dicts(places)

    with open(config["files"]["cooccurrences_path"], "rb") as f:
        cooccurrences = pickle.load(f)

    for key, order in groupby(places, lambda place: place["order_id"]):
        places_in_order = list(order)

        #print("len places_in_order:", len(places_in_order))
        #exit()

        enwiki_titles = [p["enwiki_title"] for p in places_in_order if p["enwiki_title"] and p["likely_correct"]]
        totals = defaultdict(Counter)

        for place in places_in_order:
            place_enwiki_title = place["enwiki_title"]
            place["combos"] = []
            for other_enwiki_title in enwiki_titles:
                if place_enwiki_title != other_enwiki_title:
                    key = tuple(sorted([place_enwiki_title, other_enwiki_title]))
                    value = cooccurrences.get(key, 0)
                    place["combos"].append((other_enwiki_title, value))
                    totals[place["feature_id"]][other_enwiki_title] += value

        """
        for each place in the order
        calculate the median percentage of cooccurence with likely places versus alternative features
        """
        for place in places_in_order:
            feature_totals = totals[place["feature_id"]]
            counts = [float(count) / feature_totals[title] for title, count in place["combos"] if feature_totals[title]]

            # defaults to 0.5 if none cooccurred
            place["median_cooccurrence"] = median(counts) if counts else 0.5

            # no need anymore
            del place["combos"]
        #print("places_in_order:", places_in_order)
        #exit()

    print("set cooccurrence scores")

    return places



from distutils.core import setup

setup(
  name = 'marge',
  packages = ['marge'],
  package_dir = {'marge': 'marge'},
  package_data = {'marge': ['__init__.py', 'cleaner.py', 'config.py', 'config.yml', 'converter.py', 'enricher.py', 'enumerations.py', 'models.py', 'resolver.py', 'split.py', 'train.py', 'utils.py','configs/__init__.py', 'configs/first_pass.py', 'configs/second_pass.py', 'tests/__init__.py', 'tests/cleaner.py', 'tests/models.py', 'tests/utils.py']},
  version = '0.1',
  description = 'Model that Automatically Resolves Geographic Entities',
  author = 'Daniel J. Dufour',
  author_email = 'daniel.j.dufour@gmail.com',
  url = 'https://github.com/DanielJDufour/marge',
  download_url = 'https://github.com/DanielJDufour/marge/tarball/download',
  keywords = ['location','geo','python','tensorflow'],
  classifiers = [],
  install_requires=["numpy", "pandas", "scipy"]
)



import json
import unittest

from marge.utils import *

class TestUtils(unittest.TestCase):

    def test_has_enwiki_title(self):
        train_set = to_dicts(config["files"]["dirty_training_file"], nrows=2000)
        count = 0
        for d in train_set:
            if "enwiki_title" in d:
                count += 1
            else:
                print("enwiki_title not in ", d)
        self.assertTrue(count > 100)



from csv import DictReader
from csv import DictWriter
from random import random

from config import config

input_data_file = open(config["files"]["dirty_data_file"], encoding="utf-8")
reader = DictReader(input_data_file, delimiter="\t")
fieldnames = reader.fieldnames

training = open(config["files"]["dirty_training_file"], encoding="utf-8", mode="w")
testing = open(config["files"]["dirty_testing_file"], encoding="utf-8", mode="w")

training_writer = DictWriter(training, delimiter="\t", fieldnames=fieldnames)
training_writer.writeheader()
testing_writer = DictWriter(testing, delimiter="\t", fieldnames=fieldnames)
testing_writer.writeheader()

# iterate over groups/orders
group = []
previous_group_id = None
count = 0
for row in reader:
    count +=1

    if count % 1e5 == 0:
        print("count:", count)

    group_id = row["order_id"]
    if group_id == previous_group_id:
        group.append(row)
    else:
        if group:
            n = random() >= 0.10
            if n:
                for group_row in group:
                    training_writer.writerow(group_row)
            else:
                for group_row in group:
                    testing_writer.writerow(group_row)
        group = [row]
        previous_group_id = group_id

training.close()
testing.close()



from pandas import DataFrame

# converts _input to dataframe
def to_data_frame(_input):
    if isinstance(_input, DataFrame):
        return _input
    else:
        try:
            return DataFrame(_input)
        except Exception as e:
            print("failed converting bc:", e)

# converts input into a list of dictionaries
def to_list_of_dicts(i):
    if isinstance(i, list):
        if isinstance(i[0], dict):
            return i
        else:
            print("need to iterator through and convert to dicts")
    elif isinstance(i, DataFrame):
        return i.to_dict('records')
    elif isinstance(i, str):
        if isfile(i):
            sep = "\t" if i.endswith("\t") else ","
            f = open(i)
            iterator = csv.DictReader(f, delimiter=sep)



from enricher import add_likely_correct, add_median_cooccurrence, combine_fields

from marge.config import config
from marge.utils import to_dicts
from marge.models import Model

m1 = Model("first_pass")
m1.train()
train_set = to_dicts(config["files"]["dirty_training_file"], nrows=config["training"]["rowcount"])
places = combine_fields(train_set, ["order_id", "feature_id"])
first_results = m1.predict(train_set)
print('first_results:', first_results)

places = add_likely_correct(first_results)
places = add_median_cooccurrence(places)

m2 = Model("second_pass")
print("m2:", m2)
m2.train(train_set=places)
#second_results = m2.predict(first_results)
print("m2 trained:", m2)

print("FINISHED TRAINING\n\n\n\n\n")
